# -*- coding: utf-8 -*-
"""
    weppy.tags
    ----------

    Provides html generation classes.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import re
import threading

from ._compat import iteritems, text_type, implements_to_string, to_unicode, to_native, string_types
from .libs.sanitizer import sanitize

__all__ = ['tag', 'cat', 'safe', 'asis']

THREAD_LOCAL = threading.local()


def htmlescape(s, quote=True):
    """
    returns an escaped string of the provided text s
    s: the text to be escaped
    quote: optional (default True)
    """
    # first try the xml function
    if isinstance(s, TAG):
        return s.to_html()
    # otherwise, make it a string
    if not isinstance(s, text_type):
        s = to_unicode(s)
    s = s.replace("&", "&amp;")  # Must be done first!
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    if quote:
        s = s.replace("'", "&#x27;")
        s = s.replace('"', "&quot;")
    return s


@implements_to_string
class TAG(object):
    rules = {'ul': ['li'],
             'ol': ['li'],
             'table': ['tr', 'thead', 'tbody'],
             'thead': ['tr'],
             'tbody': ['tr'],
             'tr': ['td', 'th'],
             'select': ['option', 'optgroup'],
             'optgroup': ['optionp']}
    _self_closed = ['br', 'col', 'embed', 'hr', 'img', 'input', 'link', 'meta']

    def __init__(self, name):
        self.safe = safe
        self.name = name
        self.parent = None
        self.components = []
        self.attributes = {}
        if hasattr(THREAD_LOCAL, "_tags_stack") and THREAD_LOCAL._tags_stack:
            THREAD_LOCAL._tags_stack[-1].append(self)

    def __enter__(self):
        if not hasattr(THREAD_LOCAL, "_tags_stack"):
            THREAD_LOCAL._tags_stack = []
        THREAD_LOCAL._tags_stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        THREAD_LOCAL._tags_stack.pop(-1)

    @staticmethod
    def wrap(component, rules):
        if rules and (not isinstance(component, TAG) or
                      component.name not in rules):
            return TAG(rules[0])(component)
        return component

    def __call__(self, *components, **attributes):
        rules = self.rules.get(self.name, [])
        self.components = [self.wrap(comp, rules) for comp in components]
        self.attributes = attributes
        for component in self.components:
            if isinstance(component, TAG):
                component.parent = self
        return self

    def append(self, component):
        self.components.append(component)

    def insert(self, i, component):
        self.components.insert(i, component)

    def remove(self, component):
        self.components.remove(component)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]
        else:
            return self.attributes.get(key)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.components.insert(key, value)
        else:
            self.attributes[key] = value

    def __iter__(self):
        for item in self.components:
            yield item

    def __str__(self):
        return self.to_html()

    def __add__(self, other):
        return cat(self, other)

    def add_class(self, name):
        """ add a class to _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) | set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    def remove_class(self, name):
        """ remove a class from _class attribute """
        c = self['_class']
        classes = (set(c.split()) if c else set()) - set(name.split())
        self['_class'] = ' '.join(classes) if classes else None
        return self

    regex_tag = re.compile('^([\w\-\:]+)')
    regex_id = re.compile('#([\w\-]+)')
    regex_class = re.compile('\.([\w\-]+)')
    regex_attr = re.compile('\[([\w\-\:]+)=(.*?)\]')

    def find(self, expr):
        union = lambda a, b: a.union(b)
        if ',' in expr:
            tags = reduce(union, [self.find(x.strip())
                                  for x in expr.split(',')], set())
        elif ' ' in expr:
            tags = [self]
            for k, item in enumerate(expr.split()):
                if k > 0:
                    children = [set([c for c in tag if isinstance(c, TAG)])
                                for tag in tags]
                    tags = reduce(union, children)
                tags = reduce(union, [tag.find(item) for tag in tags], set())
        else:
            tags = reduce(union, [c.find(expr)
                                  for c in self if isinstance(c, TAG)], set())
            tag = TAG.regex_tag.match(expr)
            id = TAG.regex_id.match(expr)
            _class = TAG.regex_class.match(expr)
            attr = TAG.regex_attr.match(expr)
            if (tag is None or self.name == tag.group(1)) and \
               (id is None or self['_id'] == id.group(1)) and \
               (_class is None or _class.group(1) in
                    (self['_class'] or '').split()) and \
               (attr is None or self['_' + attr.group(1)] == attr.group(2)):
                tags.add(self)
        return tags

    def _build_html_attributes(self):
        return u' '.join(
            u'%s="%s"' % (k[1:], k[1:] if v is True else htmlescape(v))
            for (k, v) in sorted(self.attributes.items())
            if k.startswith('_') and v is not None)

    def to_html(self):
        name = self.name
        ca = self._build_html_attributes()
        da = self.attributes.get('data', {})
        ca_data = u' '.join(
            u'data-%s="%s"' % (k, htmlescape(v)) for k, v in iteritems(da))
        if ca_data:
            ca = ca + u' ' + ca_data
        ca = u' ' + ca if ca else u''
        if name in self._self_closed:
            return u'<%s%s />' % (name, ca)
        else:
            co = u''.join(htmlescape(v) for v in self.components)
            return u'<%s%s>%s</%s>' % (name, ca, co, name)


class METATAG(object):
    def __getattr__(self, name):
        return TAG(name)

    def __getitem__(self, name):
        return TAG(name)

tag = METATAG()


class cat(TAG):
    def __init__(self, *components):
        self.components = [c for c in components]
        self.attributes = {}

    def to_html(self):
        return u''.join(htmlescape(v) for v in self.components)


class safe(TAG):
    default_allowed_tags = {
        'a': ['href', 'title', 'target'], 'b': [], 'blockquote': ['type'],
        'br': [], 'i': [], 'li': [], 'ol': [], 'ul': [], 'p': [], 'cite': [],
        'code': [], 'pre': [], 'img': ['src', 'alt'], 'strong': [],
        'h1': [], 'h2': [], 'h3': [], 'h4': [], 'h5': [], 'h6': [],
        'table': [], 'tr': [], 'td': ['colspan'], 'div': [],
    }

    def __init__(self, text, sanitize=False, allowed_tags=None):
        self.text = text
        self.sanitize = sanitize
        self.allowed_tags = allowed_tags or safe.default_allowed_tags

    def to_html(self):
        if self.sanitize:
            return sanitize(to_native(self.text), self.allowed_tags.keys(),
                            self.allowed_tags)
        if not isinstance(self.text, text_type):
            self.text = to_unicode(self.text)
        return self.text


def asis(text):
    return safe(text)



# -*- coding: utf-8 -*-
"""
    weppy.language.cache
    --------------------

    Adapted from the web2py's code (http://www.web2py.com)

    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>
    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

from threading import RLock
from os import stat


language_cache = {}
_cfs = {}


def get_from_cache(filename, val, fun):
    lang_dict, lock = language_cache.get(
        filename, ({}, RLock()))
    lock.acquire()
    try:
        result = lang_dict.get(val)
    finally:
        lock.release()
    if result:
        return result
    lock.acquire()
    try:
        result = lang_dict.setdefault(val, fun())
    finally:
        lock.release()
    return result


def clear_cache(filename):
    cache = language_cache.setdefault(
        filename, ({}, RLock()))
    lang_dict, lock = cache
    lock.acquire()
    try:
        lang_dict.clear()
    finally:
        lock.release()


#: returns content from filename, making sure to close the file on exit.
def _read_file(filename, mode='r'):
    with open(filename, mode) as f:
        return f.read()


#: Caches the *filtered* file `filename` with `key` until the file is modified.
def getcfs(key, filename, filter=None):
    try:
        t = stat(filename).st_mtime
    except OSError:
        return filter() if callable(filter) else ''
    lock = RLock()
    lock.acquire()
    item = _cfs.get(key, None)
    lock.release()
    if item and item[0] == t:
        return item[1]
    if not callable(filter):
        data = _read_file(filename)
    else:
        data = filter()
    lock.acquire()
    _cfs[key] = (t, data)
    lock.release()
    return data



# -*- coding: utf-8 -*-
"""
    weppy.wsgi
    ----------

    Provide error, static and dynamic handlers for wsgi.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import sys
import re
from datetime import datetime
from .http import HTTP
from .stream import stream_file_handler

REGEX_STATIC = re.compile('^/(?P<l>\w+/)?static/(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$')
REGEX_WEPPY = re.compile('^/__weppy__/(?P<f>.*?)$')
REGEX_RANGE = re.compile('^\s*(?P<start>\d*).*(?P<stop>\d*)\s*$')


def dynamic_handler(app, environ, start_response):
    try:
        #: init current
        from .globals import current
        environ["wpp.application"] = app.name
        environ["wpp.apppath"] = app.root_path
        environ["wpp.appnow"] = app.now_reference
        current.initialize(environ)
        #: dispatch request
        response = current.response
        app.route.dispatch()
        #: build HTTP response
        http = HTTP(response.status, response.output, response.headers,
                    response.cookies)
    except HTTP:
        #: catch HTTP exceptions
        http = sys.exc_info()[1]
        #: render error with handlers if in app
        error_handler = app.error_handlers.get(http.status_code)
        if error_handler:
            output = error_handler()
            http = HTTP(http.status_code, output, http.headers)
        #: store cookies
        if response.cookies:
            chead = http.cookies2header(response.cookies)
            http.headers += chead
    return http.to(environ, start_response)


def static_handler(app, environ, start_response):
    path_info = environ['PATH_INFO']
    #: handle weppy assets (helpers)
    fw_match = REGEX_WEPPY.match(path_info)
    if fw_match:
        filename = fw_match.group('f')
        static_file = os.path.join(
            os.path.dirname(__file__), 'assets', filename)
        #: avoid exposing html files
        if os.path.splitext(static_file)[1] == 'html':
            return HTTP(404).to(environ, start_response)
        return stream_file_handler(
            environ, start_response, static_file)
    #: match and process static requests
    static_match = REGEX_STATIC.match(path_info)
    if static_match:
        #: process with language urls if required
        if app.language_force_on_url:
            lang, version, filename = static_match.group('l', 'v', 'f')
            static_file = os.path.join(app.static_path, filename)
            if lang:
                lang_file = os.path.join(app.static_path, lang, filename)
                if os.path.exists(lang_file):
                    static_file = lang_file
        #: process without language urls
        else:
            version, filename = static_match.group('v', 'f')
            static_file = os.path.join(app.static_path, filename)
        return stream_file_handler(
            environ, start_response, static_file, version)
    #: process dynamic requests
    else:
        return dynamic_handler(app, environ, start_response)


def error_handler(app, environ, start_response):
    environ['wpp.now.utc'] = datetime.utcnow()
    environ['wpp.now.local'] = datetime.now()
    try:
        return static_handler(app, environ, start_response)
    except Exception:
        if app.debug:
            from .debug import smart_traceback, debug_handler
            tb = smart_traceback(app)
            body = debug_handler(tb)
        else:
            body = None
            custom_handler = app.error_handlers.get(500)
            if custom_handler:
                try:
                    body = custom_handler()
                except:
                    pass
            if not body:
                body = '<html><body>Internal error</body></html>'
        app.log.exception('Application exception:')
        return HTTP(500, body).to(environ, start_response)



# -*- coding: utf-8 -*-
"""
    weppy.dal.adapters
    ------------------

    Provides adapters facilities for dal.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from pydal.adapters import adapters
from pydal.adapters.postgres import (
    Postgre, PostgrePsyco, PostgrePG8000,
    PostgreNew, PostgrePsycoNew, PostgrePG8000New,
    PostgreBoolean, PostgrePsycoBoolean, PostgrePG8000Boolean
)


adapters._registry_.update({
    'postgres': PostgreBoolean,
    'postgres:psycopg2': PostgrePsycoBoolean,
    'postgres:pg8000': PostgrePG8000Boolean,
    'postgres2': PostgreNew,
    'postgres2:psycopg2': PostgrePsycoNew,
    'postgres2:pg8000': PostgrePG8000New,
    'postgres3': Postgre,
    'postgres3:psycopg2': PostgrePsyco,
    'postgres3:pg8000': PostgrePG8000
})



# -*- coding: utf-8 -*-
"""
    weppy.dal.migrations.helpers
    ----------------------------

    Provides helpers for migrations.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
from collections import Iterable
from uuid import uuid4
from ..._compat import string_types, integer_types
from ...templating.core import Templater
from ...datastructures import sdict, _unique_list


DEFAULT_VALUE = lambda: None


def make_migration_id():
    return uuid4().hex[-12:]


class WrappedOperation(object):
    def __init__(self, op_class, name, engine):
        self.op_class = op_class
        self.name = name
        self.engine = engine

    def __call__(self, *args, **kwargs):
        op = getattr(self.op_class, self.name)(*args, **kwargs)
        op._env_load_(self.engine)
        return op.run()


class Dispatcher(object):
    def __init__(self):
        self._registry = {}

    def dispatch_for(self, target):
        def wrap(fn):
            self._registry[target] = fn
            return fn
        return wrap

    def dispatch(self, obj):
        targets = type(obj).__mro__
        for target in targets:
            if target in self._registry:
                return self._registry[target]
        else:
            raise ValueError("no dispatch function for object: %s" % obj)


def render_template(path, template, ctx):
    _fakeapp = sdict(
        template_preloaders=[],
        template_extensions=[],
        template_lexers=[],
        debug=False,
        config=sdict(
            templates_auto_reload=False
        )
    )
    templater = Templater(_fakeapp)
    source = templater.load(path, template)
    return templater._render(source, path, template, ctx)


def to_tuple(x, default=None):
    if x is None:
        return default
    elif isinstance(x, string_types):
        return (x, )
    elif isinstance(x, Iterable):
        return tuple(x)
    else:
        return (x, )


def tuple_or_value(val):
    if not val:
        return None
    elif len(val) == 1:
        return val[0]
    else:
        return val


def tuple_rev_as_scalar(rev):
    if not rev:
        return None
    elif len(rev) == 1:
        return rev[0]
    else:
        return rev


def dedupe_tuple(tup):
    return tuple(_unique_list(tup))


def format_with_comma(value):
    if value is None:
        return ""
    elif isinstance(value, string_types):
        return value
    elif isinstance(value, Iterable):
        return ", ".join(value)
    else:
        raise ValueError("Don't know how to comma-format %r" % value)


def _feasible_as_dbms_default(val):
    if callable(val):
        return False
    if val is None:
        return True
    if isinstance(val, integer_types):
        return True
    if isinstance(val, string_types):
        return True
    if isinstance(val, (bool, float)):
        return True
    return False



# -*- coding: utf-8 -*-
"""
    weppy.tools.auth.engine
    -----------------------

    Provides the weppy authorization system.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's auth module (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import base64
import os
import time
from pydal.objects import Row
from ..._compat import iteritems, itervalues, to_unicode
from ...datastructures import sdict
from ...expose import url
from ...globals import request, session
from ...language import T
from ...http import HTTP, redirect
from ...security import uuid
from ..mail import Mail
from .defaults import default_settings, default_messages
from .exposer import Exposer
from .handlers import AuthManager, AuthLoginHandler
from .helpers import get_vars_next
from .models import AuthModel


class Auth(object):
    registered_actions = {}

    def get_or_create_key(self, app, filename=None, alg='sha512'):
        if not filename:
            path = os.path.join(app.root_path, 'private')
            if not os.path.exists(path):
                os.mkdir(path)
            filename = os.path.join(path, 'auth.key')
        else:
            filename = os.path.join(app.root_path, filename)
        if os.path.exists(filename):
            with open(filename, 'r') as f:
                key = f.read().strip()
        else:
            key = alg + ':' + uuid()
            with open(filename, 'w') as f:
                f.write(key)
        return key

    def url(self, args=[], params={}, scheme=None):
        return url(self.settings.base_url, args, params, scheme=scheme)

    def _init_settings(self, source=default_settings):
        rv = sdict()
        for key, value in iteritems(source):
            if isinstance(value, dict):
                value = self._init_settings(value)
            rv[key] = value
        return rv

    def __init__(self, app, db, usermodel=None, mailer=True, hmac_key=None,
                 hmac_key_file=None, base_url=None, csrf_prevention=True,
                 **kwargs):
        self.db = db
        self.csrf_prevention = csrf_prevention
        #: generate hmac_key if needed
        if hmac_key is None:
            hmac_key = self.get_or_create_key(app, hmac_key_file)
        #: init settings
        url_index = base_url or "account"
        settings = self.settings = self._init_settings()
        settings.update(
            base_url=url_index,
            #cas_domains=[request.env.http_host],
            #cas_provider=cas_provider,
            #cas_actions=dict(login='login',
            #                validate='validate',
            #                servicevalidate='serviceValidate',
            #                proxyvalidate='proxyValidate',
            #                logout='logout'),
            download_url="/download",
            mailer=(mailer == True) and Mail(app) or mailer,
            login_methods=[self],
            login_form=self,
            hmac_key=hmac_key
        )
        #: load user's settings
        for key, value in app.config.auth.items():
            if key in self.settings.keys():
                self.settings[key] = value
        #: load messages
        messages = self.messages = sdict()
        messages.update(default_messages)
        messages.update(self.settings.messages)
        #: init models
        self._usermodel = None
        if usermodel:
            if not issubclass(usermodel, AuthModel):
                raise RuntimeError('%s is an invalid user model' %
                                   usermodel.__name__)
            self.settings.models.user = usermodel
        self.define_models()
        #: load exposer
        self.exposer = Exposer(self)
        #: define allowed actions
        default_actions = [
            'login', 'logout', 'register', 'verify_email',
            'retrieve_username', 'retrieve_password',
            'reset_password', 'request_reset_password',
            'change_password', 'profile',
            #'groups', 'impersonate',
            'not_authorized']
        for k in default_actions:
            if k not in self.settings.actions_disabled:
                self.register_action(k, getattr(self.exposer, k))

    @property
    def _auth(self):
        return session.auth

    @property
    def user(self):
        try:
            u = session.auth.user if session.auth.user else None
        except:
            u = None
        return u

    @property
    def user_id(self):
        return self.user and self.user.id or None

    @property
    def table_user(self):
        return self.db[self._model_names.user]

    @property
    def table_group(self):
        return self.db[self._model_names.group]

    @property
    def table_membership(self):
        return self.db[self._model_names.membership]

    @property
    def table_permission(self):
        return self.db[self._model_names.permission]

    @property
    def table_event(self):
        return self.db[self._model_names.event]

    #@property
    #def table_cas(self):
    #   return self.db[self.settings.table_cas_name]

    @property
    def handler(self):
        return AuthManager(self)

    def register_action(self, name, f):
        self.registered_actions[name] = f

    def __call__(self, f, a):
        """
        usage:

        def authentication(): return dict(form=auth())
        """

        if not f:
            redirect(self.url('login', request.query_params))
        if f in self.registered_actions:
            if a is not None:
                return self.registered_actions[f](a)
            else:
                return self.registered_actions[f]()
        else:
            raise HTTP(404)

    def __get_migrate(self, tablename, migrate=True):
        if type(migrate).__name__ == 'str':
            return (migrate + tablename + '.table')
        elif migrate == False:
            return False
        else:
            return True

    def __get_modelnames(self):
        rv = {}
        def_names = {
            'user': 'user',
            'group': 'authgroup',
            'membership': 'authmembership',
            'permission': 'authpermission',
            'event': 'authevent'
        }
        for m in ['user', 'group', 'membership', 'permission', 'event']:
            if self.settings.models[m] == default_settings.models[m]:
                rv[m] = def_names[m]
            else:
                rv[m] = self.settings.models[m].__name__.lower()
        return rv

    def __set_models_labels(self):
        for model in itervalues(default_settings.models):
            for supmodel in list(reversed(model.__mro__))[1:]:
                current_labels = {}
                if hasattr(supmodel, 'form_labels'):
                    for key, val in iteritems(supmodel.form_labels):
                        current_labels[key] = T(val)
                    supmodel.form_labels = current_labels

    def define_models(self):
        self.__set_models_labels()
        names = self.__get_modelnames()
        models = self.settings.models
        #: AuthUser
        user_model = models['user']
        many_refs = [
            {names['membership'] + 's': models['membership'].__name__},
            {names['event'] + 's': models['event'].__name__},
            {names['group'] + 's': {'via': names['membership'] + 's'}},
            {names['permission'] + 's': {'via': names['group'] + 's'}}
        ]
        if getattr(user_model, '_auto_relations', True):
            for el in many_refs:
                key = list(el)[0]
                user_model._all_hasmany_ref_[key] = el
        if user_model.validation.get('password') is None:
            user_model.validation['password'] = {
                'len': {'gte': self.settings.password_min_length},
                'crypt': {'key': self.settings.hmac_key}
            }
        #: AuthGroup
        group_model = models['group']
        if not hasattr(group_model, 'format'):
            setattr(group_model, 'format', '%(role)s (%(id)s)')
        many_refs = [
            {names['membership'] + 's': models['membership'].__name__},
            {names['permission'] + 's': models['permission'].__name__},
            {names['user'] + 's': {'via': names['membership'] + 's'}}
        ]
        if getattr(group_model, '_auto_relations', True):
            for el in many_refs:
                key = list(el)[0]
                group_model._all_hasmany_ref_[key] = el
        #: AuthMembership
        membership_model = models['membership']
        belongs_refs = [
            {names['user']: models['user'].__name__},
            {names['group']: models['group'].__name__}
        ]
        if getattr(membership_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                membership_model._all_belongs_ref_[key] = el
        #: AuthPermission
        permission_model = models['permission']
        belongs_refs = [
            {names['group']: models['group'].__name__}
        ]
        if getattr(permission_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                permission_model._all_belongs_ref_[key] = el
        #: AuthEvent
        event_model = models['event']
        belongs_refs = [
            {names['user']: models['user'].__name__}
        ]
        if getattr(event_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                event_model._all_belongs_ref_[key] = el
        models.user.auth = self
        self.db.define_models(
            models.user, models.group, models.membership, models.permission,
            models.event
        )
        self._model_names = sdict()
        for key, value in iteritems(models):
            self._model_names[key] = value.__name__

    def log_event(self, description, vars={}, origin='auth'):
        """
        usage:

            auth.log_event(description='this happened', origin='auth')
        """
        if not self.settings.logging_enabled or not description:
            return
        elif self.is_logged_in():
            user_id = self.user.id
        else:
            user_id = None  # user unknown
        # log messages should not be translated
        if type(description).__name__ == 'TElement':
            description = description.m
        self.table_event.insert(
            description=str(description % vars),
            origin=origin, user=user_id)

    def get_or_create_user(self, keys, update_fields=['email'],
                           login=True, get=True):
        """
        Used for alternate login methods:
            If the user exists already then password is updated.
            If the user doesn't yet exist, then they are created.
        """
        user = None
        checks = []
        # make a guess about who this user is
        for fieldname in ['registration_id', 'username', 'email']:
            if fieldname in self.table_user.fields() and \
                    keys.get(fieldname, None):
                checks.append(fieldname)
                value = keys[fieldname]
                user = self.table_user(**{fieldname: value})
                if user:
                    break
        if not checks:
            return None
        if 'registration_id' not in keys:
            keys['registration_id'] = keys[checks[0]]
        # if we think we found the user but registration_id does not match,
        # make new user
        if 'registration_id' in checks and user and user.registration_id \
           and ('registration_id' not in keys or user.registration_id !=
                str(keys['registration_id'])):
            user = None  # THINK MORE ABOUT THIS? DO WE TRUST OPENID PROVIDER?
        if user:
            if not get:
                # added for register_bare to avoid overwriting users
                return None
            update_keys = dict(registration_id=keys['registration_id'])
            for key in update_fields:
                if key in keys:
                    update_keys[key] = keys[key]
            user.update_record(**update_keys)
        elif checks:
            if 'first_name' not in keys and 'first_name' in \
                    self.table_user.fields:
                guess = keys.get('email', 'anonymous').split('@')[0]
                keys['first_name'] = keys.get('username', guess)
            user_id = self.table_user.insert(
                **self.table_user._filter_fields(keys))
            user = self.table_user[user_id]
            if self.settings.create_user_groups:
                group_id = self.add_group(
                    self.settings.create_user_groups % user)
                self.add_membership(group_id, user_id)
            if self.settings.everybody_group_id:
                self.add_membership(self.settings.everybody_group_id, user_id)
            if login:
                self.login_user(user)
        return user

    def basic(self, basic_auth_realm=False):
        """
        perform basic login.

        :param basic_auth_realm: optional basic http authentication realm.
        :type basic_auth_realm: str or unicode or function or callable or
                                boolean.

        reads current.request.env.http_authorization
        and returns basic_allowed,basic_accepted,user.

        if basic_auth_realm is defined is a callable it's return value
        is used to set the basic authentication realm, if it's a string
        its content is used instead.  Otherwise basic authentication realm
        is set to the application name.
        If basic_auth_realm is None or False (the default) the behavior
        is to skip sending any challenge.

        """
        if not self.settings.allow_basic_login:
            return (False, False, False)
        basic = request.environ['HTTP_AUTHORIZATION']
        if basic_auth_realm:
            if callable(basic_auth_realm):
                basic_auth_realm = basic_auth_realm()
            elif basic_auth_realm is True:
                basic_realm = u'' + request.application
            basic_realm = to_unicode(basic_realm)
            http_401 = HTTP(
                401, u'Not Authorized',
                **{'WWW-Authenticate': u'Basic realm="' + basic_realm + '"'})
        if not basic or not basic[:6].lower() == 'basic ':
            if basic_auth_realm:
                raise http_401
            return (True, False, False)
        (username, sep, password) = base64.b64decode(basic[6:]).partition(':')
        is_valid_user = sep and self.login_bare(username, password)
        if not is_valid_user and basic_auth_realm:
            raise http_401
        return (True, True, is_valid_user)

    def login_user(self, user, remember=False):
        user = Row(user)
        try:
            del user.password
        except:
            pass
        expiration = remember and self.settings.long_expiration or \
            self.settings.expiration
        session.auth = sdict(
            user=user,
            last_visit=request.now,
            last_dbcheck=request.now,
            expiration=expiration,
            remember=remember,
            hmac_key=uuid())

    def login_bare(self, username, password):
        """
        logins user as specified by username (or email) and password
        """
        userfield = self.settings.login_userfield
        user = self.table_user(**{userfield: username})
        if user and user.get('password', False):
            password = self.table_user['password'].validate(
                password)[0]
            if not user.registration_key and password == user['password']:
                self.login_user(user)
                return user
        else:
            # user not in database try other login methods
            for login_method in self.settings.login_methods:
                if login_method != self and login_method(username, password):
                    self.user = username
                    return username
        return False

    def register_bare(self, **fields):
        """
        registers a user as specified by username (or email)
        and a raw password.
        """
        userfield = self.settings.login_userfield
        if not fields.get('password'):
            raise ValueError("register_bare: " +
                             "password not provided or invalid")
        elif not fields.get(userfield):
            raise ValueError("register_bare: " +
                             "userfield not provided or invalid")
        fields['password'] = self.table_user['password'].validate(
            fields['password'])[0]
        user = self.get_or_create_user(
            fields, login=False, get=False,
            update_fields=self.settings.update_fields)
        if not user:
            # get or create did not create a user (it ignores
            # duplicate records)
            return False
        return user

    def _login_with_handler(self, handler, env=None):
        if not issubclass(handler, AuthLoginHandler):
            raise RuntimeError('Provided handler for login is invalid')

        settings = self.settings
        passfield = settings.password_field
        log = self.messages['login_log']

        # redirect user if it's already logged in
        if self.user:
            redirect(self.settings.login_next or self.url('profile'))

        handler = handler(self, env)

        # use session for federated login
        snext = get_vars_next()
        if snext:
            session._auth_next = snext
        elif session._auth_next:
            snext = session._auth_next

        if handler.next is None:
            unext = settings.login_next or self.url('profile')
            if snext:
                unext = snext
        else:
            unext = handler.next

        #: init handler form if required
        #  note: we need to load the form before calling `get_user()`, as the
        #        handler could use the form itself to init the user var
        loginform = None
        if hasattr(handler, 'login_form'):
            loginform = handler.login_form()
        #: get user from handler
        user = handler.get_user()
        if user:
            if not handler.store_password:
                user[passfield] = None
            if handler.create_user_onlogin:
                user = self.get_or_create_user(
                    self.table_user._filter_fields(user),
                    settings.update_fields)
        #: return form if required
        elif loginform is not None:
            return loginform
        #: use external login url
        else:
            redirect(handler.login_url(unext))

        #: process authenticated users
        self.login_user(user, request.params.get('remember', False))
        #: log login
        self.log_event(log, user)
        #: handler callback
        handler.onsuccess()

        #: clean session next
        if unext == session._auth_next:
            del session._auth_next
        #: proceed
        redirect(unext)

    def is_logged_in(self):
        return True if self.user else False

    def email_reset_password(self, user):
        reset_password_key = str(int(time.time())) + '-' + uuid()
        link = self.url('reset_password', params={"key": reset_password_key},
                        scheme=True)
        d = dict(user)
        d.update(dict(key=reset_password_key, link=link))
        if self.settings.mailer and self.settings.mailer.send(
                to=user.email,
                subject=self.messages.reset_password_subject,
                message=self.messages.reset_password % d):
            user.update_record(reset_password_key=reset_password_key)
            return True
        return False

    def add_group(self, role, description=''):
        """
        creates a group associated to a role
        """
        group_id = self.table_group.insert(
            role=role, description=description)
        self.log_event(self.messages['add_group_log'],
                       dict(group_id=group_id, role=role))
        return group_id

    def del_group(self, group):
        """
        deletes a group
        """
        names = self.__get_modelnames()
        self.db(self.table_membership[names['group']] == group).delete()
        self.db(self.table_permission[names['group']] == group).delete()
        self.db(self.table_group.id == group).delete()
        self.log_event(self.messages.del_group_log, dict(authgroup=group))

    def id_group(self, role):
        """
        returns the group_id of the group specified by the role
        """
        rows = self.db(self.table_group.role == role).select()
        if not rows:
            return None
        return rows[0].id

    def user_group(self, user=None):
        """
        returns the group_id of the group uniquely associated to this user
        """
        return self.id_group(self.user_group_role(user))

    def user_group_role(self, user_id=None):
        if not self.settings.create_user_groups:
            return None
        if user_id:
            user = self.table_user[user_id]
        else:
            user = self.user
        return self.settings.create_user_groups % user

    def has_membership(self, group=None, user=None, role=None):
        """
        checks if user is member of group or role
        """
        names = self.__get_modelnames()
        group = group or self.id_group(role)
        try:
            group = int(group)
        except:
            group = self.id_group(group)
        if not user and self.user:
            user = self.user.id
        if group and user and self.db(
                (self.table_membership[names['user']] == user) &
                (self.table_membership[names['group']] == group)).select():
            r = True
        else:
            r = False
        self.log_event(self.messages['has_membership_log'],
                       dict(user=user, authgroup=group, check=r))
        return r

    def add_membership(self, group=None, user=None, role=None):
        """
        gives user membership of group or role
        """
        names = self.__get_modelnames()
        group = group or self.id_group(role)
        try:
            group = int(group)
        except:
            group = self.id_group(group)
        if not user and self.user:
            user = self.user.id
        data = {names['user']: user, names['group']: group}
        record = self.table_membership(**data)
        if record:
            return record.id
        else:
            rid = self.table_membership.insert(**data)
        self.log_event(self.messages['add_membership_log'],
                       dict(user=user, authgroup=group))
        return rid

    def del_membership(self, group=None, user=None, role=None):
        """
        revokes membership from group to user
        if user is None than user is that of current logged in user
        """
        names = self.__get_modelnames()
        group = group or self.id_group(role)
        if not user and self.user:
            user = self.user.id
        self.log_event(self.messages['del_membership_log'],
                       dict(user=user, authgroup=group))
        ret = self.db(
            (self.table_membership[names['user']] == user) &
            (self.table_membership[names['group']] == group)).delete()
        return ret

    def has_permission(self, name='any', table_name='', record_id=0,
                       user=None, group=None):
        """
        checks if user_id or current logged in user is member of a group
        that has 'name' permission on 'table_name' and 'record_id'
        if group_id is passed, it checks whether the group has the permission
        """
        names = self.__get_modelnames()
        query = (
            (self.table_permission.name == name) &
            (self.table_permission.table_name == table_name) &
            (self.table_permission.record_id == record_id)
        )
        if user is not None:
            parent = self.table_user(id=user)
        elif group is not None:
            parent = self.table_group(id=group)
        else:
            if not self.user:
                return False
            parent = self.user

        rows = parent[names['permission'] + 's'](query=query)
        if rows:
            return True
        return False

    def add_permission(self, group, name='any', table_name='', record_id=0):
        """
        gives group 'name' access to 'table_name' and 'record_id'
        """
        names = self.__get_modelnames()
        if group == 0:
            group = self.user_group()
        query = (
            (self.table_permission[names['group']] == group) &
            (self.table_permission.name == name) &
            (self.table_permission.table_name == str(table_name)) &
            (self.table_permission.record_id == long(record_id))
        )
        record = self.db(query).select(
            limitby=(0, 1), orderby_on_limitby=False).first()
        if record:
            rid = record.id
        else:
            data = {
                names['group']: group, 'name': name,
                'table_name': str(table_name), 'record_id': long(record_id)}
            rid = self.table_permission.insert(**data)
        self.log_event(self.messages['add_permission_log'],
                       dict(permission_id=id, group_id=group,
                            name=name, table_name=table_name,
                            record_id=record_id))
        return rid

    def del_permission(self, group, name='any', table_name='', record_id=0):
        """
        revokes group_id 'name' access to 'table_name' and 'record_id'
        """
        names = self.__get_modelnames()
        self.log_event(
            self.messages['del_permission_log'],
            dict(group_id=group, name=name, table_name=table_name,
                 record_id=record_id))
        return self.db(
            (self.table_permission[names['group']] == group) &
            (self.table_permission.name == name) &
            (self.table_permission.table_name == str(table_name)) &
            (self.table_permission.record_id == long(record_id))
        ).delete()

    def change_user_status(self, user, status):
        return self.db(self.table_user.id == user).update(
            registration_key=status)

    def disable_user(self, user):
        return self.change_user_status(user, 'disabled')

    def block_user(self, user):
        return self.change_user_status(user, 'blocked')

    def allow_user(self, user):
        return self.change_user_status(user, '')

    """
    def accessible_query(self, name, table, user_id=None):
        if not user_id:
            user_id = self.user_id
        db = self.db
        if isinstance(table, str) and table in self.db.tables():
            table = self.db[table]
        elif isinstance(table, (Set, Query)):
            # experimental: build a chained query for all tables
            if isinstance(table, Set):
                cquery = table.query
            else:
                cquery = table
            tablenames = db._adapter.tables(cquery)
            for tablename in tablenames:
                cquery &= self.accessible_query(name, tablename,
                                                user_id=user_id)
            return cquery
        if not isinstance(table, str) and\
                self.has_permission(name, table, 0, user_id):
            return table.id > 0
        query = table.id.belongs(db(self.table_membership.user_id == user_id)(
            self.table_membership.group_id == self.table_permission.group_id)(
            self.table_permission.name == name)(
            self.table_permission.table_name == table)._select(
            self.table_permission.record_id))
        if self.settings.everybody_group_id:
            query |= table.id.belongs(db(self.table_permission.group_id ==
                                         self.settings.everybody_group_id)(
                self.table_permission.name == name)(
                self.table_permission.table_name == table)._select(
                self.table_permission.record_id))
        return query

    @staticmethod
    def archive(form, archive_table=None, current_record='current_record',
                archive_current=False, fields=None):
        if not archive_current and not form.record:
            return None
        table = form.table
        if not archive_table:
            archive_table_name = '%s_archive' % table
            if not archive_table_name in table._db:
                table._db.define_table(
                    archive_table_name,
                    Field(current_record, table),
                    *[field.clone(unique=False) for field in table])
            archive_table = table._db[archive_table_name]
        new_record = {current_record: form.vars.id}
        for fieldname in archive_table.fields:
            if not fieldname in ['id', current_record]:
                if archive_current and fieldname in form.vars:
                    new_record[fieldname] = form.vars[fieldname]
                elif form.record and fieldname in form.record:
                    new_record[fieldname] = form.record[fieldname]
        if fields:
            new_record.update(fields)
        id = archive_table.insert(**new_record)
        return id
    """



# -*- coding: utf-8 -*-
"""
    tests.translator
    ----------------

    Test weppy translator module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import pytest
from weppy import App, T
from weppy.language import _instance as Tinstance
from weppy.globals import current


@pytest.fixture(scope='module')
def app():
    app = App(__name__)
    app.language_write = True
    delattr(Tinstance, '_t')
    return app


def _make_translation(language):
    return str(T('partly cloudy', language=language))


def test_translation(app):
    current._language = 'en'
    assert _make_translation('it') == 'nuvolosità variabile'
    assert _make_translation('de') == 'teilweise bewölkt'
    assert _make_translation('ru') == 'переменная облачность'


def test_write(app):
    current._language = 'en'
    #: get original content
    translated_file = os.path.join(app.root_path, 'languages', 'ru.py')
    with open(translated_file) as f:
        ocontents = f.read()
    #: write to file
    assert str(T('perché', language='ru')) == 'perché'
    #: get new contents and write back the original
    with open(translated_file) as f:
        contents = f.read()
    with open(translated_file, 'w') as f:
        f.write(ocontents)
    #: verify
    assert '"%s": "%s"' % ('perché', 'perché') in contents


# def test_cache():
#     pass



# -*- coding: utf-8 -*-
"""
    weppy.tools.auth.helpers
    ------------------------

    Provides helpers for the authorization system.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's auth module (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

from ...globals import request
from ...http import redirect


def callback(actions, form, tablename=None):
    if actions:
        if tablename and isinstance(actions, dict):
            actions = actions.get(tablename, [])
        if not isinstance(actions, (list, tuple)):
            actions = [actions]
        [action(form) for action in actions]


def call_or_redirect(f, *args):
    if callable(f):
        redirect(f(*args))
    else:
        redirect(f)


def replace_id(u, form):
    if u:
        u = u.replace('[id]', str(form.params.id))
        if u[0] == '/' or u[:4] == 'http':
            return u
    return '/account' + u


def get_vars_next():
    nextv = request.params._next
    if isinstance(nextv, (list, tuple)):
        nextv = nextv[0]
    return nextv



"""
weppy is a full-stack python web framework designed with simplicity in mind.
It takes some components from *web2py* and has a syntax inspired to *Flask*.


Links
-----

* `website <http://weppy.org>`_
* `documentation <http://weppy.org/docs>`_
* `git repo <http://github.com/gi0baro/weppy>`_

"""

import re
import ast
from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('weppy/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='weppy',
    version=version,
    url='http://github.com/gi0baro/weppy/',
    license='BSD',
    author='Giovanni Barillari',
    author_email='gi0baro@d4net.org',
    description='The web framework for humans',
    long_description=__doc__,
    packages=[
        'weppy', 'weppy.dal', 'weppy.dal.migrations', 'weppy.language',
        'weppy.language.plurals', 'weppy.templating', 'weppy.testing',
        'weppy.tools', 'weppy.tools.auth', 'weppy.validators', 'weppy.libs'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'click>=0.6',
        'pyaes',
        'pyDAL==16.9',
        'pyyaml',
        'udatetime~=0.0.9'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points={
        'console_scripts': ['weppy = weppy.cli:main']
    },
)



# -*- coding: utf-8 -*-
"""
    weppy.globals
    -------------

    Provide the current object. Used by application to deal with
    request, response, session (if loaded with handlers).

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import cgi
import copy
import json
import re
import threading

from ._compat import SimpleCookie, iteritems, to_native
from ._internal import ObjectProxy, LimitedStream
from .datastructures import sdict
from .helpers import get_flashed_messages
from .tags import htmlescape
from .utils import cachedprop
from .libs.contenttype import contenttype


_regex_client = re.compile('[\w\-:]+(\.[\w\-]+)*\.?')


class Request(object):
    def __init__(self, environ):
        self.environ = environ
        self.scheme = 'https' if \
            environ.get('wsgi.url_scheme', '').lower() == 'https' or \
            environ.get('HTTP_X_FORWARDED_PROTO', '').lower() == 'https' or \
            environ.get('HTTPS', '') == 'on' else 'http'
        self.name = '<module>.<func>'
        self.hostname = environ.get('HTTP_HOST') or '%s:%s' % \
            (environ.get('SERVER_NAME', ''), environ.get('SERVER_PORT', ''))
        self.method = environ.get('REQUEST_METHOD', 'GET').lower()
        self.path_info = environ.get('PATH_INFO') or \
            environ.get('REQUEST_URI').split('?')[0]
        self.input = environ.get('wsgi.input')
        self._now_ref = environ['wpp.appnow']
        self.nowutc = environ['wpp.now.utc']
        self.nowloc = environ['wpp.now.local']
        self.application = environ['wpp.application']

    @cachedprop
    def now(self):
        if self._now_ref == "utc":
            return self.nowutc
        return self.nowloc

    @cachedprop
    def query_params(self):
        query_string = self.environ.get('QUERY_STRING', '')
        dget = cgi.parse_qs(query_string, keep_blank_values=1)
        params = sdict(dget)
        for key, value in iteritems(params):
            if isinstance(value, list) and len(value) == 1:
                params[key] = value[0]
        return params

    def __parse_json_params(self):
        content_length = self.environ.get('CONTENT_LENGTH')
        try:
            content_length = max(0, int(content_length))
        except:
            content_length = None
        if content_length is None:
            return {}
        try:
            stream = LimitedStream(self.input, content_length)
            params = json.loads(to_native(stream.read()))
        except:
            params = {}
        return params

    @cachedprop
    def body_params(self):
        params = sdict()
        if self.environ.get('CONTENT_TYPE', '')[:16] == 'application/json':
            json_params = self.__parse_json_params()
            params.update(json_params)
            return params
        if self.input and self.environ.get('REQUEST_METHOD') in \
                ('POST', 'PUT', 'DELETE', 'BOTH'):
            dpost = cgi.FieldStorage(fp=self.input, environ=self.environ,
                                     keep_blank_values=1)
            try:
                keys = sorted(dpost)
            except:
                keys = []
            for key in keys:
                dpk = dpost[key]
                if not isinstance(dpk, list):
                    dpk = [dpk]
                dpk = [item.value if not item.filename else item
                       for item in dpk]
                params[key] = dpk
            for key, value in list(params.items()):
                if isinstance(value, list) and len(value) == 1:
                    params[key] = value[0]
        return params

    @cachedprop
    def params(self):
        rv = copy.copy(self.query_params)
        for key, val in iteritems(self.body_params):
            if key not in rv:
                rv[key] = val
            else:
                if not isinstance(rv[key], list):
                    rv[key] = [rv[key]]
                rv[key] += val if isinstance(val, list) else [val]
        return rv

    @cachedprop
    def cookies(self):
        cookies = SimpleCookie()
        for cookie in self.environ.get('HTTP_COOKIE', '').split(';'):
            cookies.load(cookie)
        return cookies

    @cachedprop
    def client(self):
        g = _regex_client.search(self.environ.get('HTTP_X_FORWARDED_FOR', ''))
        client = (g.group() or '').split(',')[0] if g else None
        if client in (None, '', 'unknown'):
            g = _regex_client.search(self.environ.get('REMOTE_ADDR', ''))
            if g:
                client = g.group()
            elif self.hostname.startswith('['):
                # IPv6
                client = '::1'
            else:
                # IPv4
                client = '127.0.0.1'
        return client

    @cachedprop
    def isajax(self):
        return self.environ.get('HTTP_X_REQUESTED_WITH', '').lower == \
            'xmlhttprequest'

    @cachedprop
    def env(self):
        #: parse the environment variables into a sdict
        _env = sdict(
            (k.lower().replace('.', '_'), v)
            for k, v in iteritems(self.environ)
        )
        return _env

    __getitem__ = object.__getattribute__
    __setitem__ = object.__setattr__


class Response(object):
    def __init__(self, environ):
        self.status = 200
        self.cookies = SimpleCookie()
        self.headers = {'Content-Type':
                        contenttype(environ['PATH_INFO'], 'text/html')}
        self.meta = sdict()
        self.meta_prop = sdict()

    def alerts(self, **kwargs):
        return get_flashed_messages(**kwargs)

    def get_meta(self):
        s = '\n'.join(
            '<meta name="%s" content="%s" />\n' % (k, htmlescape(v))
            for k, v in iteritems(self.meta or {}))
        s += '\n'.join(
            '<meta property="%s" content="%s" />\n' % (k, htmlescape(v))
            for k, v in iteritems(self.meta_prop or {}))
        return s

    __getitem__ = object.__getattribute__
    __setitem__ = object.__setattr__


class Current(threading.local):
    _language = None

    def initialize(self, environ):
        self.__dict__.clear()
        self.environ = environ
        self.request = Request(environ)
        self.response = Response(environ)
        self.session = None
        self._language = environ.get('HTTP_ACCEPT_LANGUAGE')

    @cachedprop
    def T(self):
        from .language import T
        return T


current = Current()

request = ObjectProxy(current, "request")
response = ObjectProxy(current, "response")
session = ObjectProxy(current, "session")



# -*- coding: utf-8 -*-
"""
    tests.http
    ----------------

    Test weppy http module

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from weppy.http import HTTP, redirect


def test_http_body_none():
    http = HTTP(200, body=None)

    assert http.body == []
    assert http.status_code == 200
    assert http.headers == []


def test_http_body_list():
    http = HTTP(200, body=[b'Test list'])

    assert http.body == [b'Test list']
    assert http.status_code == 200
    assert http.headers == []


def test_http():
    response = []
    buffer = []

    def start_response(status, headers):
        response[:] = [status, headers]
        return buffer.append

    http = HTTP(200, 'Hello World',
                headers={'X-Test': 'Hello Header'},
                cookies={'cookie_test': 'hello cookie'})

    assert http.body == [b'Hello World']
    assert http.status_code == 200
    assert http.headers == [('X-Test', 'Hello Header'), ('Set-Cookie', 'e')]

    assert http.to({'REQUEST_METHOD': 'HEAD'}, start_response) == [b'']
    assert http.to({'REQUEST_METHOD': 'GET'}, start_response) == [b'Hello World']


def test_redirect():
    from weppy.globals import current
    current.initialize({
        'PATH_INFO': '/',
        'wpp.appnow': 'test',
        'wpp.now.utc': 'test',
        'wpp.now.local': 'test',
        'wpp.application': 'test',
    })

    try:
        redirect('/redirect', 302)
    except HTTP as http_redirect:
        assert current.response.status == 302
        assert http_redirect.status_code == 302
        assert http_redirect.headers == [('Location', '/redirect')]



# -*- coding: utf-8 -*-
"""
    weppy.sessions
    --------------

    Provides session handlers for weppy applications.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import time
import tempfile
from ._compat import pickle, to_native
from .security import secure_loads, secure_dumps, uuid
from .handlers import Handler
from .globals import current, request, response
from .datastructures import sdict, SessionData


class SessionCookieManager(Handler):
    def __init__(self, key, secure=False, domain=None):
        self.key = key
        self.secure = secure
        self.domain = domain

    def on_start(self):
        self.cookie_data_name = 'wpp_session_data_%s' % request.application
        if self.cookie_data_name in request.cookies:
            cookie_data = request.cookies[self.cookie_data_name].value
            current.session = SessionData(secure_loads(
                cookie_data, self.key), expires=3600)
        if not current.session:
            current.session = SessionData(expires=3600)

    def on_end(self):
        data = secure_dumps(sdict(current.session), self.key)
        response.cookies[self.cookie_data_name] = data
        response.cookies[self.cookie_data_name]['path'] = "/"
        response.cookies[self.cookie_data_name]['expires'] = \
            current.session._expiration
        if self.secure:
            response.cookies[self.cookie_data_name]['secure'] = True
        if self.domain is not None:
            response.cookies[self.cookie_data_name]['domain'] = self.domain

    def clear(self):
        raise NotImplementedError(
            "%s doesn't support clear of sessions. " +
            "You should change the '%s' parameter to invalidate existing ones."
            % (self.__class__.__name__, 'secure'))


class SessionFSManager(Handler):
    _fs_transaction_suffix = '.__wp_sess'
    _fs_mode = 0o600

    def __init__(self, expire=3600, secure=False, domain=None,
                 filename_template='weppy_%s.sess'):
        assert not filename_template.endswith(self._fs_transaction_suffix), \
            'filename templates cannot end with %s' % \
            self._fs_transaction_suffix
        self._filename_template = filename_template
        from .expose import Expose
        self._path = os.path.join(Expose.application.root_path, 'sessions')
        #: create required paths if needed
        if not os.path.exists(self._path):
            os.mkdir(self._path)
        self.expire = expire
        self.secure = secure
        self.domain = domain

    def _get_filename(self, sid):
        sid = to_native(sid)
        return os.path.join(self._path, self._filename_template % sid)

    def _load(self, sid):
        try:
            f = open(self._get_filename(sid), 'rb')
        except IOError:
            return None
        now = time.time()
        exp = pickle.load(f)
        val = pickle.load(f)
        f.close()
        if exp < now:
            f.close()
            return None
        return val

    def _store(self, session, expiration):
        fn = self._get_filename(session._sid)
        now = time.time()
        exp = now + expiration
        fd, tmp = tempfile.mkstemp(suffix=self._fs_transaction_suffix,
                                   dir=self._path)
        f = os.fdopen(fd, 'wb')
        try:
            pickle.dump(exp, f, 1)
            #f.write(session._dump)
            pickle.dump(sdict(session), f, pickle.HIGHEST_PROTOCOL)
        finally:
            f.close()
        try:
            os.rename(tmp, fn)
            os.chmod(fn, self._fs_mode)
        except:
            pass

    def _delete(self, session):
        fn = self._get_filename(session._sid)
        try:
            os.unlink(fn)
        except OSError:
            pass

    def on_start(self):
        self.cookie_data_name = 'wpp_session_data_%s' % request.application
        if self.cookie_data_name in request.cookies:
            sid = request.cookies[self.cookie_data_name].value
            data = self._load(sid)
            if data is not None:
                current.session = SessionData(data, sid=sid)
        if not current.session:
            sid = uuid()
            current.session = SessionData(sid=sid)

    def on_end(self):
        if not current.session:
            self._delete(current.session)
            if current.session._modified:
                #: if we got here means we want to destroy session definitely
                if self.cookie_data_name in response.cookies:
                    del response.cookies[self.cookie_data_name]
            return
        #: store and update cookies
        expiration = current.session._expiration or self.expire
        if current.session._modified:
            self._store(current.session, expiration)
        response.cookies[self.cookie_data_name] = current.session._sid
        response.cookies[self.cookie_data_name]['path'] = "/"
        response.cookies[self.cookie_data_name]['expires'] = expiration
        if self.secure:
            response.cookies[self.cookie_data_name]['secure'] = True
        if self.domain is not None:
            response.cookies[self.cookie_data_name]['domain'] = self.domain

    def clear(self):
        for element in os.listdir(self._path):
            os.unlink(os.path.join(self._path, element))


class SessionRedisManager(Handler):
    def __init__(self, redis, prefix="wppsess:", expire=3600, secure=False,
                 domain=None):
        self.redis = redis
        self.prefix = prefix
        self.expire = expire
        self.secure = secure
        self.domain = domain

    def on_start(self):
        self.cookie_data_name = 'wpp_session_data_%s' % request.application
        if self.cookie_data_name in request.cookies:
            sid = request.cookies[self.cookie_data_name].value
            #: load from redis
            data = self.redis.get(self.prefix + sid)
            if data is not None:
                current.session = SessionData(pickle.loads(data), sid=sid)
        if not current.session:
            sid = uuid()
            current.session = SessionData(sid=sid)

    def on_end(self):
        if not current.session:
            self.redis.delete(self.prefix + current.session._sid)
            if current.session._modified:
                #: if we got here means we want to destroy session definitely
                if self.cookie_data_name in response.cookies:
                    del response.cookies[self.cookie_data_name]
            return
        #: store on redis and update cookies
        expiration = current.session._expiration or self.expire
        if current.session._modified:
            self.redis.setex(self.prefix + current.session._sid,
                             current.session._dump, expiration)
        else:
            self.redis.expire(self.prefix + current.session._sid, expiration)
        response.cookies[self.cookie_data_name] = current.session._sid
        response.cookies[self.cookie_data_name]['path'] = "/"
        response.cookies[self.cookie_data_name]['expires'] = expiration
        if self.secure:
            response.cookies[self.cookie_data_name]['secure'] = True
        if self.domain is not None:
            response.cookies[self.cookie_data_name]['domain'] = self.domain

    def clear(self):
        self.redis.delete(self.prefix + "*")



# -*- coding: utf-8 -*-
"""
    weppy.stream
    ------------

    Provide files streaming interface.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import re
import time
import errno

from .http import HTTP
from .libs.contenttype import contenttype

__all__ = ['stream_file_handler']

REGEX_STATIC = re.compile('^/(?P<a>.*?)/static/(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$')
REGEX_RANGE = re.compile('^\s*(?P<start>\d*).*(?P<stop>\d*)\s*$')


class FileSubset(object):
    """ class needed to handle RANGE currents """
    def __init__(self, stream, start, stop):
        self.stream = stream
        self.stream.seek(start)
        self.size = stop - start

    def read(self, bytes=None):
        bytes = self.size if bytes is None else max(bytes, self.size)
        if bytes:
            data = self.stream.read(bytes)
            self.size -= bytes
            return data
        else:
            return ''

    def close(self):
        self.stream.close()


def streamer(environ, static_file, version=None, headers=None,
             block_size=10**5):
    try:
        stream = open(static_file, 'rb')
        headers = headers or dict()
        fsize = os.path.getsize(static_file)
        modified = os.path.getmtime(static_file)
        mtime = time.strftime(
            '%a, %d %b %Y %H:%M:%S GMT', time.gmtime(modified))
        headers['Content-Type'] = contenttype(static_file)
        # check if file has to be served as an attachment
        if environ.get('QUERY_STRING', '').startswith('attachment'):
            headers['Content-Disposition'] = 'attachment; filename="%s"' % \
                static_file
        # check if file modified since or not
        if environ.get('HTTP_IF_MODIFIED_SINCE') == mtime:
            return HTTP(304, headers=headers)
        headers['Last-Modified'] = mtime
        headers['Pragma'] = 'cache'
        if version:
            headers['Cache-Control'] = 'max-age=315360000'
            headers['Expires'] = 'Thu, 31 Dec 2037 23:59:59 GMT'
        else:
            headers['Cache-Control'] = 'private'
        # check whether a range request and serve partial content accordingly
        http_range = environ.get('HTTP_RANGE', None)
        if http_range:
            status = 206
            match = REGEX_RANGE.match(http_range)
            start = match.group('start') or 0
            stop = match.group('stop') or (fsize - 1)
            stream = FileSubset(stream, start, stop + 1)
            headers['Content-Range'] = 'bytes %i-%i/%i' % (start, stop, fsize)
            headers['Content-Length'] = '%i' % (stop - start + 1)
        else:
            status = 200
            if 'gzip' in environ.get('HTTP_ACCEPT_ENCODING', ''):
                gzipped = static_file + '.gz'
                if os.path.isfile(gzipped) and \
                   os.path.getmtime(gzipped) > modified:
                    stream.close()
                    static_file = gzipped
                    fsize = os.path.getsize(gzipped)
                    headers['Content-Encoding'] = 'gzip'
                    headers['Vary'] = 'Accept-Encoding'
                    stream = open(static_file, 'rb')
            headers['Content-Length'] = str(fsize)
    except IOError as e:
        if e.errno in (errno.EISDIR, errno.EACCES):
            return HTTP(403, body="FORBIDDEN\n")
        return HTTP(404, body="NOT FOUND\n")
    else:
        # serve using wsgi.file_wrapper is available
        if 'wsgi.file_wrapper' in environ:
            data = environ['wsgi.file_wrapper'](stream, block_size)
        else:
            data = iter(lambda: stream.read(block_size), '')
        return HTTP(status, data, headers=headers)


def stream_file_handler(environ, start_response, static_file, version=None,
                        headers=None, block_size=10**5):
    return streamer(environ, static_file, version, headers, block_size).to(
        environ, start_response)



# -*- coding: utf-8 -*-
"""
    weppy.dal.helpers
    -----------------

    Provides helpers for dal.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import re
import time
from collections import OrderedDict
from pydal._globals import THREAD_LOCAL
from pydal.objects import Query, IterRows as _IterRows
from pydal.helpers.classes import Reference as _IDReference, ExecutionHandler
from .._compat import implements_iterator, iteritems
from ..datastructures import sdict
from ..utils import cachedprop
from .base import Set, Field, Rows


class Reference(object):
    def __init__(self, *args, **params):
        self.reference = [arg for arg in args]
        self.params = params
        self.refobj[id(self)] = self

    def __call__(self, func):
        if self.__class__.__name__ not in ['has_one', 'has_many']:
            raise SyntaxError(
                '%s cannot be used as a decorator' % self.__class__.__name__)
        if not callable(func):
            raise SyntaxError('Argument must be callable')
        if self.reference:
            raise SyntaxError(
                "When using %s as decorator, you must use the 'field' option" %
                self.__class__.__name__)
        new_reference = {func.__name__: {'method': func}}
        field = self.params.get('field')
        if field:
            new_reference[func.__name__]['field'] = field
        self.reference = [new_reference]
        return self

    @property
    def refobj(self):
        return {}


class ReferenceData(sdict):
    def __init__(self, model_class, **kwargs):
        self.model_class = model_class
        super(ReferenceData, self).__init__(**kwargs)

    @cachedprop
    def dbset(self):
        if self.method:
            return self.method(self.model_class)
        return self.model_class.db

    @cachedprop
    def model_instance(self):
        if self.method:
            return self.dbset._model_
        return self.dbset[self.model]._model_

    @property
    def table(self):
        return self.model_instance.table

    @property
    def table_name(self):
        return self.model_instance.tablename

    @property
    def field_instance(self):
        return self.table[self.field]


class RelationBuilder(object):
    def __init__(self, ref, model):
        self.ref = ref
        self.model = model

    def _make_refid(self, row):
        return row.id if row is not None else self.model.id

    def _extra_scopes(self, ref, model_instance=None):
        model_instance = model_instance or ref.model_instance
        rv = []
        if ref.scope is not None:
            scope_m = model_instance._scopes_[ref.scope].f
            rv.append(lambda f=scope_m, m=model_instance: f(m))
        if ref.where is not None:
            rv.append(lambda f=ref.where, m=model_instance: f(m))
        return rv

    def _patch_query_with_scopes(self, ref, query):
        for scope in self._extra_scopes(ref):
            query = query & scope()
        return query

    def _patch_query_with_scopes_on(self, ref, query, model_name):
        model = self.model.db[model_name]._model_
        for scope in self._extra_scopes(ref, model):
            query = query & scope()
        return query

    def _get_belongs(self, modelname, value):
        return self.model.db[modelname]._model_._belongs_ref_.get(value)

    def belongs_query(self):
        return (self.model.table[self.ref[1]] == self.model.db[self.ref[0]].id)

    @staticmethod
    def many_query(ref, rid):
        return ref.model_instance.table[ref.field] == rid

    def _many(self, ref, rid):
        return ref.dbset.where(
            self._patch_query_with_scopes(ref, self.many_query(ref, rid))
        ).query

    def many(self, row=None):
        return self._many(self.ref, self._make_refid(row))

    def via(self, row=None):
        db = self.model.db
        rid = self._make_refid(row)
        sname = self.model.__class__.__name__
        stack = []
        midrel = self.model._hasmany_ref_[self.ref.via]
        stack.append(self.ref)
        while midrel.via is not None:
            stack.insert(0, midrel)
            midrel = self.model._hasmany_ref_[midrel.via]
        query = self._many(midrel, rid)
        step_model = midrel.table_name
        sel_field = db[step_model].ALL
        last_belongs = None
        last_via = None
        for via in stack:
            rname = via.field or via.name[:-1]
            belongs_model = self._get_belongs(step_model, rname)
            if belongs_model:
                #: join table way
                last_belongs = step_model
                last_via = via
                _query = (db[belongs_model].id == db[step_model][rname])
                sel_field = db[belongs_model].ALL
                step_model = belongs_model
            else:
                #: shortcut way
                last_belongs = None
                rname = via.field or via.name
                midrel = db[step_model]._model_._hasmany_ref_[rname]
                _query = self._many(midrel, db[step_model].id)
                step_model = midrel.table_name
                sel_field = db[step_model].ALL
            query = query & _query
        query = via.dbset.where(
            self._patch_query_with_scopes_on(via, query, step_model)).query
        return query, sel_field, sname, rid, last_belongs, last_via


class RelationSet(object):
    _relation_method_ = 'many'

    def __init__(self, db, relation_builder, row):
        self.db = db
        self._relation_ = relation_builder
        self._row_ = row

    def _get_query_(self):
        try:
            return getattr(self._relation_, self._relation_method_)(self._row_)
        except AttributeError as e:
            raise RuntimeError(e.message)
        except Exception:
            raise

    @property
    def _model_(self):
        return self._relation_.ref.model_instance

    @cachedprop
    def _field_(self):
        return self._relation_.ref.field_instance

    @cachedprop
    def _scopes_(self):
        if self._relation_.ref.method:
            return [lambda: self._relation_.ref.dbset.query]
        return self._relation_._extra_scopes(self._relation_.ref)

    @cachedprop
    def _set(self):
        return Set(self.db, self._get_query_(), model=self._model_)

    def __getattr__(self, name):
        return getattr(self._set, name)

    def __repr__(self):
        return repr(self._set)

    def __call__(self, query, ignore_common_filters=False):
        return self._set.where(query, ignore_common_filters)

    def select(self, *args, **kwargs):
        if 'reload' in kwargs:
            del kwargs['reload']
        return self._set.select(*args, **kwargs)

    def create(self, **kwargs):
        attributes = self._get_fields_from_scopes(
            self._scopes_, self._model_.tablename)
        attributes.update(**kwargs)
        attributes[self._field_.name] = self._row_.id
        return self._model_.create(
            **attributes
        )

    @staticmethod
    def _get_fields_from_scopes(scopes, table_name):
        rv = {}
        for scope in scopes:
            query = scope()
            components = [query.second, query.first]
            current_kv = []
            while components:
                component = components.pop()
                if isinstance(component, Query):
                    components.append(component.second)
                    components.append(component.first)
                else:
                    if isinstance(component, Field) and \
                       component._tablename == table_name:
                        current_kv.append(component)
                    else:
                        if current_kv:
                            current_kv.append(component)
                        else:
                            components.pop()
                if len(current_kv) > 1:
                    rv[current_kv[0].name] = current_kv[1]
                    current_kv = []
        return rv


class HasOneSet(RelationSet):
    @cachedprop
    def _last_resultset(self):
        return self.select().first()

    def __call__(self, *args, **kwargs):
        if not args and not kwargs:
            return self._last_resultset
        return self.select(*args, **kwargs).first()


class HasOneWrap(object):
    def __init__(self, ref):
        self.ref = ref

    def __call__(self, model, row):
        return HasOneSet(model.db, RelationBuilder(self.ref, model), row)


class HasManySet(RelationSet):
    @cachedprop
    def _last_resultset(self):
        return self.select()

    def __call__(self, *args, **kwargs):
        if not args and not kwargs:
            return self._last_resultset
        return self.select(*args, **kwargs)

    def add(self, obj):
        attributes = self._get_fields_from_scopes(
            self._scopes_, self._model_.tablename)
        attributes[self._field_.name] = self._row_.id
        return self.db(
            self.db[self._field_._tablename].id == obj.id
        ).validate_and_update(**attributes)

    def remove(self, obj):
        if self.db[self._field_._tablename][self._field_.name]._isrefers:
            return self.db(
                self._field_._table.id == obj.id).validate_and_update(
                **{self._field_.name: None}
            )
        return self.db(self._field_._table.id == obj.id).delete()


class HasManyWrap(object):
    def __init__(self, ref):
        self.ref = ref

    def __call__(self, model, row):
        return HasManySet(model.db, RelationBuilder(self.ref, model), row)


class HasManyViaSet(RelationSet):
    _relation_method_ = 'via'
    _via_error = "Can't %s elements in has_many relations without a join table"

    @cachedprop
    def _viadata(self):
        query, rfield, model_name, rid, via, viadata = \
            super(HasManyViaSet, self)._get_query_()
        return sdict(
            query=query, rfield=rfield, model_name=model_name, rid=rid,
            via=via, data=viadata
        )

    def _get_query_(self):
        return self._viadata.query

    @property
    def _model_(self):
        return self.db[self._viadata.model_name]._model_

    @cachedprop
    def _last_resultset(self):
        return self.select(self._viadata.rfield)

    def __call__(self, *args, **kwargs):
        if not args:
            if not kwargs:
                return self._last_resultset
            args = [self._viadata.rfield]
        return self.select(*args, **kwargs)

    def _get_relation_fields(self):
        viadata = self._viadata.data
        self_field = self._model_._hasmany_ref_[viadata.via].field
        rel_field = viadata.field or viadata.name[:-1]
        return self_field, rel_field

    def _fields_from_scopes(self):
        viadata = self._viadata.data
        rel = self._model_._hasmany_ref_[viadata.via]
        if rel.method:
            scopes = [lambda: rel.dbset.query]
        else:
            scopes = self._relation_._extra_scopes(rel)
        return self._get_fields_from_scopes(scopes, rel.table_name)

    def create(self, **kwargs):
        raise RuntimeError('Cannot create third objects for many relations')

    def add(self, obj, **kwargs):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'add')
        nrow = self._fields_from_scopes()
        nrow.update(**kwargs)
        #: get belongs references
        self_field, rel_field = self._get_relation_fields()
        nrow[self_field] = self._viadata.rid
        nrow[rel_field] = obj.id
        #: validate and insert
        return self.db[self._viadata.via]._model_.create(nrow)

    def remove(self, obj):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'remove')
        #: get belongs references
        self_field, rel_field = self._get_relation_fields()
        #: delete
        return self.db(
            (self.db[self._viadata.via][self_field] == self._viadata.rid) &
            (self.db[self._viadata.via][rel_field] == obj.id)).delete()


class HasManyViaWrap(object):
    def __init__(self, ref):
        self.ref = ref

    def __call__(self, model, row):
        return HasManyViaSet(model.db, RelationBuilder(self.ref, model), row)


class VirtualWrap(object):
    def __init__(self, model, virtual):
        self.model = model
        self.virtual = virtual

    def __call__(self, row, *args, **kwargs):
        if self.virtual.inject_model:
            row = row[self.model.tablename]
        return self.virtual.f(self.model, row, *args, **kwargs)


class ScopeWrap(object):
    def __init__(self, set, model, scope):
        self.set = set
        self.model = model
        self.scope = scope

    def __call__(self, *args, **kwargs):
        return self.set.where(
            self.scope(self.model, *args, **kwargs), model=self.model)


class Callback(object):
    _inst_count_ = 0

    def __init__(self, f, t):
        self.t = []
        if isinstance(f, Callback):
            self.t += f.t
            f = f.f
        self.f = f
        self.t.append(t)
        self._inst_count_ = Callback._inst_count_
        Callback._inst_count_ += 1

    def __call__(self):
        return None


class JoinedIDReference(_IDReference):
    @classmethod
    def _from_record(cls, record, table=None):
        rv = cls(record.id)
        rv._table = table
        rv._record = record
        return rv

    def as_dict(self, datetime_to_str=False, custom_types=None):
        return self._record.as_dict()


class JoinableSet(Set):
    def _iterselect_rows(self, *fields, **attributes):
        tablenames = self.db._adapter.tables(
            self.query, attributes.get('join', None),
            attributes.get('left', None), attributes.get('orderby', None),
            attributes.get('groupby', None))
        fields = self.db._adapter.expand_all(fields, tablenames)
        colnames, sql = self.db._adapter._select_wcols(
            self.query, fields, **attributes)
        return JoinIterRows(self.db, sql, fields, colnames)

    def _split_joins(self, joins):
        one, many = [], []
        for jname, jtable, is_single_row in joins:
            if is_single_row:
                one.append((jname, jtable))
            else:
                many.append((jname, jtable))
        return one, many

    def _build_merged_row(self, row, many_joins):
        for jname, jtable in many_joins:
            row[jname] = OrderedDict()
        return row

    def _build_records_from_joined(self, joined, joins, many_joins, colnames):
        for rid, row in iteritems(joined):
            for jname, jtable in many_joins:
                row[jname] = Rows(self.db, list(row[jname].values()), [])
        return JoinRows(
            self.db, list(joined.values()), colnames, jtables=joins)


class JoinSet(JoinableSet):
    @classmethod
    def _from_set(cls, obj, table, joins):
        rv = cls(
            obj.db, obj.query, obj.query.ignore_common_filters)
        rv._stable_ = table
        rv._joins_ = joins
        return rv

    def select(self, *fields, **options):
        #: use iterselect for performance
        rows = self._iterselect_rows(*fields, **options)
        #: rebuild rowset using nested objects
        rowmap = OrderedDict()
        one_joins, many_joins = self._split_joins(self._joins_)
        for row in rows:
            rid = row[self._stable_].id
            rowmap[rid] = rowmap.get(
                rid, self._build_merged_row(row[self._stable_], many_joins)
            )
            for jname, jtable in one_joins:
                rowmap[rid][jname] = JoinedIDReference._from_record(
                    row[jtable], self.db[jtable])
            for jname, jtable in many_joins:
                rowmap[rid][jname][row[jtable].id] = rowmap[rid][jname].get(
                    row[jtable].id, row[jtable])
        return self._build_records_from_joined(
            rowmap, self._joins_, many_joins, rows.colnames)


class LeftJoinSet(JoinableSet):
    @classmethod
    def _from_set(cls, obj, jdata):
        rv = cls(
            obj.db, obj.query, obj.query.ignore_common_filters, obj._model_)
        rv._stable_ = rv._model_.tablename
        rv._jdata_ = jdata
        return rv

    def select(self, *fields, **options):
        #: collect tablenames
        jtables, one_joins, many_joins = [], [], []
        for index, join in enumerate(options['left']):
            jname, is_single_row = self._jdata_[index]
            tname = join.first._tablename
            jtables.append((jname, tname, is_single_row))
            if is_single_row:
                one_joins.append((jname, tname))
            else:
                many_joins.append((jname, tname))
        #: use iterselect for performance
        rows = self._iterselect_rows(*fields, **options)
        #: rebuild rowset using nested objects
        rowmap = OrderedDict()
        for row in rows:
            rid = row[self._stable_].id
            rowmap[rid] = rowmap.get(
                rid, self._build_merged_row(row[self._stable_], many_joins)
            )
            for jname, jtable in one_joins:
                if row[jtable].id:
                    rowmap[rid][jname] = JoinedIDReference._from_record(
                        row[jtable], self.db[jtable])
            for jname, jtable in many_joins:
                if row[jtable].id:
                    rowmap[rid][jname][row[jtable].id] = \
                        rowmap[rid][jname].get(row[jtable].id, row[jtable])
        return self._build_records_from_joined(
            rowmap, jtables, many_joins, rows.colnames)


@implements_iterator
class JoinIterRows(_IterRows):
    def __init__(self, db, sql, fields, colnames):
        self.db = db
        self.fields = fields
        self.colnames = colnames
        (self.fields_virtual, self.fields_lazy, self.tmps) = \
            self.db._adapter._parse_expand_colnames(colnames)
        self.cursor = self.db._adapter.cursor
        self.db._adapter.execute(sql)
        self.db._adapter.lock_cursor(self.cursor)
        self._head = None
        self.last_item = None
        self.last_item_id = None
        self.blob_decode = True
        self.cacheable = False
        self.sql = sql

    def __next__(self):
        db_row = self.cursor.fetchone()
        if db_row is None:
            raise StopIteration
        return self.db._adapter._parse(
            db_row, self.tmps, self.fields, self.colnames, self.blob_decode,
            self.cacheable, self.fields_virtual, self.fields_lazy)


class JoinRows(Rows):
    def __init__(self, *args, **kwargs):
        self._joins_ = kwargs['jtables']
        del kwargs['jtables']
        super(JoinRows, self).__init__(*args, **kwargs)

    def as_list(self, compact=True, storage_to_dict=True,
                datetime_to_str=False, custom_types=None):
        if storage_to_dict:
            items = []
            for row in self:
                item = row.as_dict(datetime_to_str, custom_types)
                for jdata in self._joins_:
                    if not jdata[2]:
                        item[jdata[0]] = row[jdata[0]].as_list()
                items.append(item)
        else:
            items = [item for item in self]
        return items


class TimingHandler(ExecutionHandler):
    def _timings(self):
        THREAD_LOCAL._weppydal_timings_ = getattr(
            THREAD_LOCAL, '_weppydal_timings_', [])
        return THREAD_LOCAL._weppydal_timings_

    @cachedprop
    def timings(self):
        return self._timings()

    def before_execute(self, command):
        self.t = time.time()

    def after_execute(self, command):
        dt = time.time() - self.t
        self.timings.append((command, dt))


def make_tablename(classname):
    words = re.findall('[A-Z][^A-Z]*', classname)
    tablename = '_'.join(words)
    return tablename.lower() + "s"



"""
    weppy.language
    --------------

    Provides the languages translator system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

from ..utils import cachedprop


class Instance(object):
    @cachedprop
    def _t(self):
        from ..expose import Expose
        from .translator import Translator
        return Translator(Expose.application)

    def T(self, *args, **kwargs):
        return self._t(*args, **kwargs)

# We use a 'proxied' object to the translator to avoid errors
# when user calls T() outside the request flow.
_instance = Instance()
T = _instance.T



# -*- coding: utf-8 -*-
"""
    weppy.templating.contents
    -------------------------

    Provides structures for templating system.

    :copyright: (c) 2014-2016 by Giovanni Barillari

    Based on the web2py's templating system (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: BSD, see LICENSE for more details.
"""

from .._compat import implements_to_string, to_unicode


def _gen_output(node, blocks):
    if isinstance(node, BlockNode):
        if node.name in blocks:
            rv = blocks[node.name].output(blocks)
        else:
            rv = node.output(blocks)
    else:
        rv = to_unicode(node)
    return rv


@implements_to_string
class Node(object):
    """
    Basic Container Object
    """
    def __init__(self, value=None, pre_extend=False, template=None,
                 lines=None):
        self.value = value
        self.pre_extend = pre_extend
        self.template = template
        self.lines = lines or (None, None)

    def __str__(self):
        return to_unicode(self.value)

    def _rendered_lines(self):
        return str(self).split("\n")[1:]


@implements_to_string
class SuperNode(Node):
    def __init__(self, name='', pre_extend=False):
        self.name = name
        self.value = None
        self.pre_extend = pre_extend

    def __str__(self):
        if self.value:
            return to_unicode(self.value)
        else:
            return u''

    def __repr__(self):
        return "%s->%s" % (self.name, self.value)


@implements_to_string
class BlockNode(Node):
    """
    Block Container.

    This Node can contain other Nodes and will render in a hierarchical order
    of when nodes were added.

    ie::

        {{ block test }}
            This is default block test
        {{ end }}
    """
    def __init__(self, name='', pre_extend=False, delimiters=('{{', '}}')):
        """
        name - Name of this Node.
        """
        self.nodes = []
        self.name = name
        self.pre_extend = pre_extend
        self.left, self.right = delimiters

    def __repr__(self):
        lines = ['%sblock %s%s' % (self.left, self.name, self.right)]
        lines += [str(node) for node in self.nodes]
        lines.append('%send%s' % (self.left, self.right))
        return ''.join(lines)

    def __str__(self):
        """
        Get this BlockNodes content, not including child Nodes
        """
        return u''.join(str(node) for node in self.nodes
                        if not isinstance(node, BlockNode))

    def append(self, node):
        """
        Add an element to the nodes.

        Keyword Arguments

        - node -- Node object or string to append.
        """
        if isinstance(node, (str, Node)):
            self.nodes.append(node)
        else:
            raise TypeError("Invalid type; must be instance of ``str`` or " +
                            "``BlockNode``. %s" % node)

    def extend(self, other):
        """
        Extend the list of nodes with another BlockNode class.

        Keyword Arguments

        - other -- BlockNode or Content object to extend from.
        """
        if isinstance(other, BlockNode):
            self.nodes.extend(other.nodes)
        else:
            raise TypeError(
                "Invalid type; must be instance of ``BlockNode``. %s" % other)

    def output(self, blocks):
        """
        Merges all nodes into a single string.
        blocks -- Dictionary of blocks that are extending
        from this template.
        """
        return u''.join(_gen_output(node, blocks) for node in self.nodes)


@implements_to_string
class Content(BlockNode):
    """
    Parent Container -- Used as the root level BlockNode.

    Contains functions that operate as such.
    """
    def __init__(self, name="ContentBlock", pre_extend=False):
        """
        Keyword Arguments

        name -- Unique name for this BlockNode
        """
        self.name = name
        self.nodes = []
        self.blocks = {}
        self.pre_extend = pre_extend
        self.template = name

    def __repr__(self):
        return object.__repr__(self)

    def __str__(self):
        return u''.join(_gen_output(node, self.blocks) for node in self.nodes)

    def _insert(self, other, index=0):
        """
        Inserts object at index.
        """
        if isinstance(other, (str, Node)):
            self.nodes.insert(index, other)
        else:
            raise TypeError(
                "Invalid type, must be instance of ``str`` or ``Node``.")

    def insert(self, other, index=0):
        """
        Inserts object at index.

        You may pass a list of objects and have them inserted.
        """
        if isinstance(other, (list, tuple)):
            # Must reverse so the order stays the same.
            other.reverse()
            for item in other:
                self._insert(item, index)
        else:
            self._insert(other, index)

    def append(self, node):
        """
        Adds a node to list. If is a BlockNode then we assign a block for it.
        """
        if isinstance(node, (str, Node)):
            self.nodes.append(node)
            if isinstance(node, BlockNode):
                self.blocks[node.name] = node
        else:
            raise TypeError("Invalid type, must be instance of ``str`` or " +
                            "``BlockNode``. %s" % node)

    def extend(self, other):
        """
        Extends the objects list of nodes with another objects nodes
        """
        if isinstance(other, BlockNode):
            self.nodes.extend(other.nodes)
            self.blocks.update(other.blocks)
        else:
            raise TypeError(
                "Invalid type; must be instance of ``BlockNode``. %s" % other)

    def clear_content(self):
        self.nodes = []



# -*- coding: utf-8 -*-
"""
    weppy.templating.helpers
    ------------------------

    Provides helpers for templating system.

    :copyright: (c) 2014-2016 by Giovanni Barillari
    :license: BSD, see LICENSE for more details.
"""

import os
import traceback
from .contents import BlockNode


class TemplateMissingError(Exception):
    def __init__(self, tpath, filename):
        self.path = tpath
        self.template = filename
        message = "Template %s not found" % self.template
        Exception.__init__(self, message)


class TemplateError(Exception):
    def __init__(self, tpath, message, filename, lineno):
        Exception.__init__(self, message)
        self.path = tpath
        self.template = filename
        if isinstance(lineno, tuple):
            lineno = lineno[0]
        self.lineno = lineno

    @property
    def file_path(self):
        return os.path.join(self.path, self.template)


class TemplateReference(object):
    def __init__(self, parserdata, code, exc_type, exc_value, tb):
        self.parser = parserdata
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.tb = tb
        if hasattr(exc_value, 'lineno'):
            dummy_lineno = exc_value.lineno
        else:
            template_frame = traceback.extract_tb(tb, 2)[-1]
            dummy_lineno = template_frame[1]
        self.lines = self.get_template_reference(parserdata.content,
                                                 parserdata.blocks)
        self.template, self.lineno = self.match_template(dummy_lineno)

    @property
    def file_path(self):
        return os.path.join(self.parser.path, self.template)

    @property
    def message(self):
        location = 'File "%s", line %d' % (self.file_path, self.lineno)
        lines = [self.args[0], '  ' + location]
        return "\n".join(lines)

    def __str__(self):
        return str(self.exc_value)

    @staticmethod
    def get_template_reference(content, blocks):
        lines = []
        for node in content.nodes:
            if isinstance(node, BlockNode):
                if node.name in blocks:
                    lines += TemplateReference.get_template_reference(
                        blocks[node.name], blocks)
                else:
                    lines += TemplateReference.get_template_reference(node,
                                                                      blocks)
            else:
                node_lines = node._rendered_lines()
                if len(node_lines) == node.lines[1] - node.lines[0] + 1:
                    linenos = [(i, i + 1) for i in range(node.lines[0],
                                                         node.lines[1] + 1)]
                else:
                    linenos = [(node.lines[0], node.lines[1])
                               for i in range(0, len(node_lines))]
                for l in range(0, len(node_lines)):
                    lines.append((node.template, linenos[l]))
        return lines

    def match_template(self, dummy_lineno):
        try:
            reference = self.lines[dummy_lineno - 1]
        except:
            reference = (self.parser.name, ('<unknown>', 'unknown'))
        return reference[0], reference[1][0]



# -*- coding: utf-8 -*-
"""
::

    # from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496942
    # Title: Cross-site scripting (XSS) defense
    # Submitter: Josh Goldfoot (other recipes)
    # Last Updated: 2006/08/05
    # Version no: 1.0

"""

from weppy._compat import PY2
if PY2:
    from htmllib import HTMLParser
    from urlparse import urlparse
    from htmlentitydefs import entitydefs
else:
    from html.parser import HTMLParser
    from urllib.parse import urlparse
    from html.entities import entitydefs
from cgi import escape
from formatter import AbstractFormatter

from xml.sax.saxutils import quoteattr

__all__ = ['sanitize']


def xssescape(text):
    """Gets rid of < and > and & and, for good measure, :"""

    return escape(text, quote=True).replace(':', '&#58;')


class XssCleaner(HTMLParser):

    def __init__(
        self,
        permitted_tags=['a', 'b', 'blockquote', 'br/', 'i',
                        'li', 'ol', 'ul', 'p', 'cite',
                        'code', 'pre', 'img/', ],
        allowed_attributes={
            'a': ['href', 'title'],
            'img': ['src', 'alt'],
            'blockquote': ['type']},
        fmt=AbstractFormatter,
        strip_disallowed=False
    ):

        HTMLParser.__init__(self, fmt)
        self.result = ''
        self.open_tags = []
        self.permitted_tags = [i for i in permitted_tags if i[-1] != '/']
        self.requires_no_close = [i[:-1] for i in permitted_tags
                                  if i[-1] == '/']
        self.permitted_tags += self.requires_no_close
        self.allowed_attributes = allowed_attributes

        # The only schemes allowed in URLs (for href and src attributes).
        # Adding "javascript" or "vbscript" to this list would not be smart.

        self.allowed_schemes = ['http', 'https', 'ftp']

        #to strip or escape disallowed tags?
        self.strip_disallowed = strip_disallowed
        self.in_disallowed = False

    def handle_data(self, data):
        if data and not self.in_disallowed:
            self.result += xssescape(data)

    def handle_charref(self, ref):
        if self.in_disallowed:
            return
        elif len(ref) < 7 and ref.isdigit():
            self.result += '&#%s;' % ref
        else:
            self.result += xssescape('&#%s' % ref)

    def handle_entityref(self, ref):
        if self.in_disallowed:
            return
        elif ref in entitydefs:
            self.result += '&%s;' % ref
        else:
            self.result += xssescape('&%s' % ref)

    def handle_comment(self, comment):
        if self.in_disallowed:
            return
        elif comment:
            self.result += xssescape('<!--%s-->' % comment)

    def handle_starttag(
        self,
        tag,
        method,
        attrs,
    ):
        if tag not in self.permitted_tags:
            if self.strip_disallowed:
                self.in_disallowed = True
            else:
                self.result += xssescape('<%s>' % tag)
        else:
            bt = '<' + tag
            if tag in self.allowed_attributes:
                attrs = dict(attrs)
                self.allowed_attributes_here = [x for x in
                                                self.allowed_attributes[tag] if x in attrs and
                                                len(attrs[x]) > 0]
                for attribute in self.allowed_attributes_here:
                    if attribute in ['href', 'src', 'background']:
                        if self.url_is_acceptable(attrs[attribute]):
                            bt += ' %s="%s"' % (attribute,
                                                attrs[attribute])
                    else:
                        bt += ' %s=%s' % (xssescape(attribute),
                                          quoteattr(attrs[attribute]))
            if bt == '<a' or bt == '<img':
                return
            if tag in self.requires_no_close:
                bt += ' /'
            bt += '>'
            self.result += bt
            if tag not in self.requires_no_close:
                self.open_tags.insert(0, tag)

    def handle_endtag(self, tag, attrs):
        bracketed = '</%s>' % tag
        if tag not in self.permitted_tags:
            if self.strip_disallowed:
                self.in_disallowed = False
            else:
                self.result += xssescape(bracketed)
        elif tag in self.open_tags:
            self.result += bracketed
            self.open_tags.remove(tag)

    def unknown_starttag(self, tag, attributes):
        self.handle_starttag(tag, None, attributes)

    def unknown_endtag(self, tag):
        self.handle_endtag(tag, None)

    def url_is_acceptable(self, url):
        """
        Accepts relative and absolute urls
        """

        parsed = urlparse(url)
        return (parsed[0] in self.allowed_schemes and '.' in parsed[1]) \
            or (parsed[0] == '' and parsed[2].startswith('/'))

    def strip(self, rawstring, escape=True):
        """
        Returns the argument stripped of potentially harmful
        HTML or Javascript code

        @type escape: boolean
        @param escape: If True (default) it escapes the potentially harmful
          content, otherwise remove it
        """

        if not isinstance(rawstring, str):
            return str(rawstring)
        for tag in self.requires_no_close:
            rawstring = rawstring.replace("<%s/>" % tag, "<%s />" % tag)
        if not escape:
            self.strip_disallowed = True
        self.result = ''
        self.feed(rawstring)
        for endtag in self.open_tags:
            if endtag not in self.requires_no_close:
                self.result += '</%s>' % endtag
        return self.result

    def xtags(self):
        """
        Returns a printable string informing the user which tags are allowed
        """

        tg = ''
        for x in sorted(self.permitted_tags):
            tg += '<' + x
            if x in self.allowed_attributes:
                for y in self.allowed_attributes[x]:
                    tg += ' %s=""' % y
            tg += '> '
        return xssescape(tg.strip())


def sanitize(
    text,
    permitted_tags=['a', 'b', 'blockquote', 'br/', 'i', 'li',
                    'ol', 'ul', 'p', 'cite', 'code', 'pre',
                    'img/', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
                    'table', 'tr', 'td', 'div', ],
    allowed_attributes={
        'a': ['href', 'title'],
        'img': ['src', 'alt'],
        'blockquote': ['type'],
        'td': ['colspan'],
    },
    escape=True
):
    if not isinstance(text, basestring):
        return str(text)
    return XssCleaner(permitted_tags=permitted_tags,
                      allowed_attributes=allowed_attributes).strip(text, escape)
